import hmac
import hashlib


def digest_size(params) -> int:
    hasher = hashlib.new(params["digest"])
    return hasher.digest_size


def sign_bytes(in_bytes: bytes, in_key: bytes, params) -> bytes:
    return hmac.digest(msg=in_bytes, key=in_key, **params)


def derive_key(**kwargs) -> bytes:
    return hashlib.scrypt(**kwargs)


def xor_bytes(in_bytes: bytes, in_key: bytes) -> bytes:
    """xors the in_bytes with a sequence of bytes with in_key"""
    return bytes(_d ^ _k for (_d, _k) in zip(in_bytes, in_key, strict=True))


def merge_data_and_signature(in_data: bytes, in_signature: bytes) -> bytes:
    return in_data + in_signature


def split_data_and_signature(
    in_bytes: bytes, signature_size: int
) -> tuple[bytes, bytes]:
    if len(in_bytes) < signature_size:
        raise ValueError("in_bytes is shorter than signature_size")
    data = in_bytes[:-signature_size]
    signature = in_bytes[-signature_size:]
    return data, signature



from .simple import simple
from .simple import spiced


def get_configurator(**kwargs):
    encryption = kwargs.get("encryption", "spiced")
    configurators = {
        "simple": simple.Simple,
        "spiced": spiced.Spiced,
    }
    new_kwargs = {_k: _v for _k, _v in kwargs.items() if _k != "encryption"}
    return configurators[encryption](**new_kwargs)



import inspect
import textwrap
import typing
import importlib.metadata

import black

from .puzzle_data_encryption import encrypt_data
from .puzzle_data_creators import question_answer_list_to_dict
from .configurators import configurators
from . import rp_configurators
from . import bytestr_utils


def _advertisement() -> str:
    package_name = "puzzle-generator"
    full_name = f"{package_name} {importlib.metadata.version(package_name)}"
    return f"""# generated with {full_name}
#
# https://pypi.org/project/puzzle-generator/
# https://github.com/vil02/puzzle_generator/
"""


def _str_to_code(in_str: str, max_len: int, quotes: str) -> str:
    return "\n".join(
        quotes + line + quotes
        for line in textwrap.wrap(
            in_str,
            width=max_len,
            replace_whitespace=False,  # Preserves embedded whitespace like `\n`
            drop_whitespace=False,  # Keeps leading/trailing spaces
        )
    )


def _create_str(in_encrypted_puzzle, configurator, rp_configurator) -> str:
    modules: str = "\n".join("import " + _ for _ in configurator.get_modules()) + "\n"
    objects: str = "\n".join(
        inspect.getsource(_)
        for _ in configurator.get_needed_objects()
        + rp_configurator.get_needed_objects()
    )
    question = _str_to_code(in_encrypted_puzzle[0], 78, '"""')
    rest_str = _str_to_code(
        bytestr_utils.bytes_to_bytestr(in_encrypted_puzzle[1]), 78, '"'
    )

    return (
        "\n".join(
            [
                _advertisement(),
                modules,
                'BYTEORDER: typing.Literal["little", "big"] = "little"',
                objects,
                rp_configurator.puzzle_data(question, rest_str),
                configurator.get_constants_str(),
                rp_configurator.call(),
            ]
        )
        + "\n"
    )


def extract_qa_list_and_hints(
    puzzle_description: typing.Sequence[
        tuple[str, str, typing.Callable[[str], str] | None] | str
    ],
) -> tuple[list[str], list[typing.Callable[[str], str] | None]]:
    qa_list: list[str] = []
    hints: list[typing.Callable[[str], str] | None] = []
    for _ in puzzle_description[:-1]:
        question, answer, hint = _
        qa_list += [question, answer]
        hints.append(hint)
    if not isinstance(puzzle_description[-1], str):
        raise ValueError(
            "In case of puzzle with hints, "
            "the last entry of the puzzle_description must be a string"
        )
    qa_list.append(puzzle_description[-1])
    return qa_list, hints


def create(
    puzzle_description: (
        list[str]
        | typing.Sequence[tuple[str, str, typing.Callable[[str], str] | None] | str]
    ),
    **kwargs,
) -> str:
    if isinstance(puzzle_description, list) and all(
        isinstance(_, str) for _ in puzzle_description
    ):
        qa_list: list[str] = puzzle_description
        hints: list[typing.Callable[[str], str] | None] = [
            None for _ in range(len(qa_list) // 2 + 1)
        ]
    else:
        qa_list, hints = extract_qa_list_and_hints(puzzle_description)

    puzzle = question_answer_list_to_dict(qa_list)
    configurator = configurators.get_configurator(**kwargs)
    encrypted_puzzle = encrypt_data(puzzle, configurator.get_encrypt())
    res = _create_str(
        encrypted_puzzle, configurator, rp_configurators.get_rp_configurator(hints)
    )
    return black.format_str(res, mode=black.FileMode())



import secrets

from ...encryption_algorithms.simple import eas_common
from ... import bytestr_utils as bsu
from ... import bytes_utils as bu
from ... import puzzle_data_encryption as pde

MODULES = ["hmac", "hashlib", "base64", "sys", "typing"]

OBJECTS = [
    eas_common.sign_bytes,
    eas_common.derive_key,
    eas_common.split_data_and_signature,
    eas_common.digest_size,
    eas_common.xor_bytes,
    bsu.bytestr_to_bytes,
    bu.bytes_to_int,
    bu.split,
    pde.decrypt_data,
]


def scrypt_params(**kwargs):
    in_params = kwargs.get("scrypt_params", {})
    default = {"salt": secrets.token_bytes(16), "n": 2**14, "r": 8, "p": 1}
    if "maxmem" in in_params:
        default["maxmem"] = in_params["maxmem"]
    res = {_k: in_params.get(_k, _v) for _k, _v in default.items()}
    if "maxmem" not in res:
        res["maxmem"] = (128 + 100) * res["n"] * res["r"] * res["p"]
    return res


def signature_params(**kwargs):
    return kwargs.get("signature_params", {"digest": "sha512"})


def scrypt_params_to_code_str(**kwargs) -> str:
    pieces = [f'"{_k}":{_v}' for _k, _v in kwargs.items() if _k != "salt"]
    salt_str = '"' + bsu.bytes_to_bytestr(kwargs["salt"]) + '"'
    pieces.append(f'"salt":bytestr_to_bytes({salt_str})')
    return f"_SCRYPT_PARAMS = {{{','.join(pieces)}}}"



def question_answer_list_to_dict(qa_list: list[str]):
    if len(qa_list) % 2 == 0:
        raise ValueError("The question/answer list must have odd length.")
    if len(qa_list) == 1:
        return {"str": qa_list[0]}
    return {
        "str": qa_list[0],
        "pass": qa_list[1],
        "rest": question_answer_list_to_dict(qa_list[2:]),
    }



import typing


from .eas_common import (
    derive_key,
    xor_bytes,
    sign_bytes,
    digest_size,
    merge_data_and_signature,
    split_data_and_signature,
)


def get_encrypt(
    scrypt_params, signature_params
) -> typing.Callable[[bytes, bytes], bytes]:
    def _encrypt(in_bytes: bytes, in_pass: bytes) -> bytes:
        signature = sign_bytes(in_bytes, in_pass, signature_params)
        merged = merge_data_and_signature(in_bytes, signature)
        key = derive_key(password=in_pass, dklen=len(merged), **scrypt_params)
        return xor_bytes(merged, key)

    return _encrypt


def get_decrypt(
    scrypt_params, signature_params
) -> typing.Callable[[bytes, bytes], bytes | None]:
    def _decrypt(in_bytes: bytes, in_pass: bytes) -> bytes | None:
        key = derive_key(password=in_pass, dklen=len(in_bytes), **scrypt_params)
        data = xor_bytes(in_bytes, key)
        decrypted, signature = split_data_and_signature(
            data, digest_size(signature_params)
        )

        if sign_bytes(decrypted, in_pass, signature_params) == signature:
            return decrypted
        return None

    return _decrypt



import typing

from . import run_puzzle as rp


class _NoHintsConfigurator:
    def get_needed_objects(self):
        return [rp.run_puzzle]

    def puzzle_data(self, question: str, rest_str: str) -> str:
        return f"_PUZZLE = ({question}, bytestr_to_bytes({rest_str}))"

    def call(self) -> str:
        return "run_puzzle(_PUZZLE, _DECRYPT, input)"


def get_proc_answer(
    in_hints: typing.Sequence[typing.Callable[[str], str] | None],
) -> typing.Callable[[str], str]:
    cur_answer = 0

    def _proc_answer(in_str: str) -> str:
        nonlocal cur_answer
        res = in_str
        cur_hint = in_hints[cur_answer]
        cur_answer += 1
        if cur_hint is not None:
            res = cur_hint(res)
        return res

    return _proc_answer


def _get_name(in_fun: typing.Callable[[str], str] | None) -> str:
    if in_fun is None:
        return "None"
    return in_fun.__name__


class _WithHintsConfigurator:
    def __init__(self, in_hints: list[typing.Callable[[str], str] | None]):
        self.hints = in_hints

    def get_needed_objects(self):
        return [get_proc_answer, rp.run_puzzle, rp.run_puzzle_with_hints] + [
            _ for _ in set(self.hints) if _ is not None
        ]

    def puzzle_data(self, question: str, rest_str: str) -> str:
        hints_str = ", ".join(_get_name(_) for _ in self.hints)
        hints = f"[{hints_str}]"
        return f"_HINTS = {hints}\n_PUZZLE = ({question}, bytestr_to_bytes({rest_str}))"

    def call(self) -> str:
        return (
            "run_puzzle_with_hints(_PUZZLE, _DECRYPT, input, get_proc_answer(_HINTS))"
        )


def get_rp_configurator(hints: list[None | typing.Callable[[str], str]]):
    if all(_ is None for _ in hints):
        return _NoHintsConfigurator()
    return _WithHintsConfigurator(hints)
